import heaobject.root
from .. import response
from ..heaobjectsupport import new_heaobject_from_type, has_permissions, RESTPermissionGroup
from ..appproperty import HEA_DB, HEA_CACHE
from .mongo import MongoContext
from heaobject.error import DeserializeException
from aiohttp.web import Request, StreamResponse, Response
from typing import Any, AsyncGenerator, Literal, IO
from heaobject.root import DesktopObject, DesktopObjectDict, desktop_object_from_dict
from heaobject.user import NONE_USER
from heaserver.service.oidcclaimhdrs import SUB
from pymongo.errors import WriteError
from collections.abc import Sequence, Mapping


async def get_dict(request: Request, collection: str,
                           volume_id: str | None = None) -> DesktopObjectDict | None:
    """
    Gets the requested desktop object as a desktop object dict.

    :param request: the HTTP request, which must have an id value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :return: a desktop object dict or None if the object was not found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    cache_key = (sub, f'id^{request.match_info["id"]}')
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return cached_value
    else:
        async with MongoContext(request, volume_id) as mongo:
            result = await mongo.get(request, collection, var_parts='id',
                                    sub=request.headers.get(SUB))

            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                permitted = has_permissions(obj=obj, sub=request.headers.get(SUB),
                                            permissions=RESTPermissionGroup.GETTER_PERMS)
                if not permitted:
                    return None
                request.app[HEA_CACHE][cache_key] = result
            return result


async def get(request: Request, collection: str, volume_id: str | None = None) -> Response:
    """
    Gets the desktop object with the specified id. The desktop object is
    formatted according to the requested mime types in the HTTP request's
    Accept header as one of the formats supported by the
    heaserver.service.representor module.

    :param request: the HTTP request, which must have an id value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response with the requested HEA object or Not Found.
    """
    result = await get_dict(request, collection, volume_id)
    return await response.get(request, desktop_object_from_dict(result).to_dict() if result is not None else None)


async def get_content(request: Request, collection: str, volume_id: str | None = None) -> StreamResponse:
    """
    Gets the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: an aiohttp StreamResponse with the requested HEA object or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        result = await mongo.get(request, collection, var_parts='id', sub=request.headers.get(SUB))

        if result is not None:
            obj = heaobject.root.desktop_object_from_dict(result)
            permitted = has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.GETTER_PERMS)
            if not permitted:
                return response.status_not_found()
        out = await mongo.get_content(request, collection, var_parts='id', sub=request.headers.get(SUB))
        if out is not None:
            return await response.get_streaming(request, out, 'text/plain')
        else:
            return response.status_not_found()


async def get_by_name(request: Request, collection: str,
                      volume_id: str | None = None) -> Response:
    """
    Gets an HTTP response object with the requested desktop object in the body.
    The desktop object is formatted according to the requested mime types in
    the HTTP request's Accept header as one of the formats supported by the
    heaserver.service.representor module.

    :param request: the HTTP request, which must have a name value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested
    desktop object. If None, the root volume is assumed.
    :return: a Response with the requested desktop object or Not Found.
    """
    result = await get_by_name_dict(request, collection, volume_id)
    return await response.get(request, desktop_object_from_dict(result).to_dict() if result is not None else None)

async def get_by_name_dict(request: Request, collection: str,
                           volume_id: str | None = None) -> DesktopObjectDict | None:
    """
    Gets the requested desktop object as a desktop object dict.

    :param request: the HTTP request, which must have a name value in the match_info mapping. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :return: a desktop object dict or None if the object was not found.
    """
    sub = request.headers.get(SUB, NONE_USER)
    cache_key = (sub, f'name^{request.match_info["name"]}')
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return cached_value
    else:
        async with MongoContext(request, volume_id) as mongo:
            result = await mongo.get(request, collection, var_parts='name',
                                    sub=request.headers.get(SUB))

            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                permitted = has_permissions(obj=obj, sub=request.headers.get(SUB),
                                            permissions=RESTPermissionGroup.GETTER_PERMS)
                if not permitted:
                    return None
                request.app[HEA_CACHE][cache_key] = result
            return result


async def get_all(request: Request,
                  collection: str,
                  volume_id: str | None = None,
                  mongoattributes: Any | None = None,
                  sort: dict[str, Literal[-1, 1]] | None = None) -> Response:
    """
    Gets an HTTP response with all requested desktop objects in the body.
    The desktop objects are formatted according to the requested mime types in
    the HTTP request's Accept header as one of the formats supported by the
    heaserver.service.representor module.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :param sort: the sort order of the desktop objects as a dict of desktop
    object attribute to 1 or -1 for ascending or descending.
    :return: a Response with a list of HEA object dicts. If no desktop objects
    are found, the body will contain an empty list.
    """
    sub = request.headers.get(SUB, None)
    cache_key = (sub, None, sort)
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return await response.get_all(request, cached_value)
    else:
        l: list[DesktopObjectDict] = []
        async for d in get_all_gen(request, collection, volume_id, mongoattributes, sort):
            l.append(desktop_object_from_dict(d).to_dict())
        request.app[HEA_CACHE][cache_key] = l
        return await response.get_all(request, l)

async def get_all_dict(request: Request,
                       collection: str,
                       volume_id: str | None = None,
                       mongoattributes: Any | None = None,
                       sort: dict[str, Literal[-1, 1]] | None = None) -> list[DesktopObjectDict]:
    """
    Gets all HEA objects as a list of desktop object dicts.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :param sort: the sort order of the desktop objects as a dict of desktop
    object attribute to 1 or -1 for ascending or descending.
    :return: a list of DesktopObjectDict. If no desktop objects are found, the
    return value will be an empty list.
    """
    sub = request.headers.get(SUB, None)
    cache_key = (sub, None, sort)
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if cached_value is not None:
        return await response.get_all(request, cached_value)
    else:
        l: list[DesktopObjectDict] = []
        async for desktop_object_dict in get_all_gen(request, collection,
                                                volume_id=volume_id,
                                                mongoattributes=mongoattributes,
                                                sort=sort):
            l.append(desktop_object_dict)
        request.app[HEA_CACHE][cache_key] = l
        return l

async def get_all_gen(request: Request,
                      collection: str,
                      volume_id: str | None = None,
                      mongoattributes: Any | None = None,
                      sort: dict[str, Literal[-1, 1]] | None = None) -> AsyncGenerator[DesktopObjectDict, None]:
    """
    Gets an async generator of all HEA objects as desktop object dicts.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA
    object. If None, the root volume is assumed.
    :param sort: the sort order of the desktop objects as a dict of desktop
    object attribute to 1 or -1 for ascending or descending.
    :return: an async generator of DesktopObjectDicts.
    """
    sub = request.headers.get(SUB)
    async with MongoContext(request, volume_id) as mongo:
        async for r in mongo.get_all(request, collection, mongoattributes=mongoattributes, sub=sub, sort=sort):
            obj = heaobject.root.desktop_object_from_dict(r)
            if has_permissions(obj=obj,
                            sub=sub,
                            permissions=RESTPermissionGroup.GETTER_PERMS):
                yield r


async def opener(request: Request, collection: str, volume_id: str | None = None,
                 include_desktop_object: bool = True) -> Response:
    """
    Gets choices for opening an HEA desktop object's content.

    :param request: the HTTP request. Required. If an Accepts header is provided, MIME types that do not support links
    will be ignored.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with status code 300, and a body containing the HEA desktop object and links
    representing possible choices for opening the HEA desktop object; or Not Found.
    """
    sub = request.headers.get(SUB, None)
    cache_key = (sub, f"id^{request.match_info['id']}")
    cached_value = request.app[HEA_CACHE].get(cache_key)
    if include_desktop_object is None:
        include_desktop_object_ = True
    else:
        include_desktop_object_ = bool(include_desktop_object)
    if cached_value is not None:
        return await response.get_multiple_choices(request, cached_value if include_desktop_object_ else None)
    else:
        async with MongoContext(request, volume_id) as mongo:
            result = await mongo.get(request, collection, var_parts='id', sub=request.headers.get(SUB))
            if result is not None:
                obj = heaobject.root.desktop_object_from_dict(result)
                permitted = has_permissions(obj=obj, sub=request.headers.get(SUB, NONE_USER), permissions=RESTPermissionGroup.GETTER_PERMS)
                if not permitted:
                    return response.status_not_found()
            else:
                return response.status_not_found()
            request.app[HEA_CACHE][cache_key] = result
            return await response.get_multiple_choices(request, result if include_desktop_object_ else None)


async def post(request: Request, collection: str, type_: type[DesktopObject], default_content: IO | None = None, volume_id: str | None = None) -> Response:
    """
    Posts the provided HEA object.

    :param request: the HTTP request.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param default_content: an optional blank document or other default content as a file-like object. This must be not-None
    for any microservices that manage content.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with a status of Created and the object's URI in the
    """
    sub = request.headers.get(SUB, None)
    if not await request.app[HEA_DB].is_creator(request, type_):
        return response.status_forbidden(f'Permission denied creating objects of type {type_.get_type_name()}')
    try:
        obj = await new_heaobject_from_type(request, type_)
    except DeserializeException as e:
        return response.status_bad_request(str(e))
    if obj.owner is None:
        obj.owner = request.headers.get(SUB, None)
    async with MongoContext(request, volume_id) as mongo:
        result = await mongo.post(request, obj, collection, default_content)
        to_delete = []
        for cache_key in request.app[HEA_CACHE]:
            if cache_key[1] is None:
                to_delete.append(cache_key)
        for cache_key in to_delete:
            request.app[HEA_CACHE].pop(cache_key, None)
        return await response.post(request, result, collection)

async def post_dict_return_id(request: Request, obj_dict: DesktopObjectDict, collection: str, type_: type[DesktopObject], default_content: IO | None = None, volume_id: str | None = None) -> str | None:
    try:
        obj = desktop_object_from_dict(obj_dict)
    except (DeserializeException, ValueError, TypeError) as e:
        return response.status_bad_request(str(e))
    if obj.owner is None:
        obj.owner = request.headers.get(SUB, None)
    async with MongoContext(request, volume_id) as mongo:
        return await mongo.post(request, obj, collection, default_content)

async def post_dict(request: Request, obj_dict: DesktopObjectDict, collection: str, type_: type[DesktopObject], default_content: IO | None = None, volume_id: str | None = None) -> Response:
    result = post_dict_return_id(request, obj_dict, collection, type_, default_content, volume_id)
    return await response.post(request, result, collection)

async def put(request: Request, collection: str, type_: type[DesktopObject], volume_id: str | None = None) -> Response:
    """
    Updates the HEA object with the specified id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    sub = request.headers.get(SUB, None)
    async with MongoContext(request, volume_id) as mongo:
        try:
            obj = await new_heaobject_from_type(request, type_)
        except DeserializeException as e:
            return response.status_bad_request(str(e).encode())
        try:
            permitted = has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.PUTTER_PERMS)
            if not permitted:
                if has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.GETTER_PERMS):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()

            result = await mongo.put(request, obj, collection, sub=sub)  # if lacks permissions or object is not in database, then updates no records.
        except WriteError as e:
            if e.code == 66:
                return response.status_bad_request(e.details['errmsg'])
            else:
                return response.status_internal_error(e.details.get('errmsg'))
        if result is not None and result.matched_count:
            to_delete = []
            for cache_key in request.app[HEA_CACHE]:
                if cache_key[1] in (None, f"id^{request.match_info['id']}"):
                    to_delete.append(cache_key)
            for cache_key in to_delete:
                request.app[HEA_CACHE].pop(cache_key, None)
        return await response.put(result.matched_count if result else False)

async def upsert(request: Request, collection: str, type_: type[DesktopObject], volume_id: str | None = None, filter: dict[str, Any] = None) -> Response:
    """
    Updates the HEA object, using the specified filter if provided otherwise the object's id, and inserting a new object
    if none matches the filter or the id.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :param filter: optional filter criteria.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    sub = request.headers.get(SUB, None)
    async with MongoContext(request, volume_id) as mongo:
        try:
            obj = await new_heaobject_from_type(request, type_)
        except DeserializeException as e:
            return response.status_bad_request(str(e).encode())
        try:
            permitted = has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.PUTTER_PERMS) and \
                has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.POSTER_PERMS)
            if not permitted:
                if has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.GETTER_PERMS):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()

            result = await mongo.upsert(request, obj, collection, sub=sub, mongoattributes=filter)
        except WriteError as e:
            if e.code == 66:
                return response.status_bad_request(e.details['errmsg'])
            else:
                return response.status_internal_error(e.details.get('errmsg'))
        to_delete = []
        for cache_key in request.app[HEA_CACHE]:
            if cache_key[1] in (None, f"id^{request.match_info['id']}"):
                to_delete.append(cache_key)
        for cache_key in to_delete:
            request.app[HEA_CACHE].pop(cache_key, None)
        return await response.put(result.matched_count if result else False)


async def put_content(request: Request, collection: str, type_: type[DesktopObject], volume_id: str | None = None) -> Response:
    """
    Updates the HEA object's associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param type_: The HEA object type. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: a Response object with a status of No Content, Forbidden, or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        sub = request.headers.get(SUB)
        result = await mongo.get(request, collection, var_parts='id', sub=sub)

        if result is not None:
            obj = heaobject.root.desktop_object_from_dict(result)
            permitted = has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.PUTTER_PERMS)
            if not permitted:
                if has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.GETTER_PERMS):
                    return response.status_forbidden()
                else:
                    return response.status_not_found()
            result2 = await mongo.put_content(request, collection, sub=sub)  # if lacks permissions, then updates no records.
            return await response.put(result2)
        else:
            return response.status_not_found()


async def delete(request: Request, collection: str, volume_id: str | None = None) -> Response:
    """
    Deletes the HEA object with the specified id and any associated content.

    :param request: the HTTP request. Required.
    :param collection: the Mongo collection name. Required.
    :param volume_id: the id string of the volume containing the requested HEA object. If None, the root volume is
    assumed.
    :return: No Content, Forbidden, or Not Found.
    """
    async with MongoContext(request, volume_id) as mongo:
        sub = request.headers.get(SUB)
        result = await mongo.get(request, collection, var_parts='id', sub=request.headers.get(SUB))

        if result is not None:
            obj = heaobject.root.desktop_object_from_dict(result)
            permitted = has_permissions(obj=obj, sub=request.headers.get(SUB), permissions=RESTPermissionGroup.DELETER_PERMS)
            if not permitted:
                return response.status_forbidden()
            result = await mongo.delete(request, collection, var_parts='id', sub=sub)  # if lacks permissions, then deletes no records.
            if result.deleted_count:
                to_delete = []
                for cache_key in request.app[HEA_CACHE]:
                    if cache_key[1] in (None, f"id^{request.match_info['id']}"):
                        to_delete.append(cache_key)
                for cache_key in to_delete:
                    request.app[HEA_CACHE].pop(cache_key, None)
        else:
            return response.status_not_found()
        return await response.delete(result.deleted_count)


async def ping(request: Request) -> Response:
    """
    Sends a ping command to the database.

    :param request: the HTTP request.
    :return: an HTTP response with status code 200 if the ping is successful, 500 otherwise.
    """
    async with MongoContext(request) as mongo:
        try:
            await mongo.ping()
            return response.status_ok()
        except Exception as e:  # The exact exception is not documented.
            raise response.status_internal_error() from e

async def aggregate(request: Request, collection: str, pipeline: Sequence[Mapping[str, Any]], volume_id: str | None = None) -> Response:
    """
    Execute an aggregation pipeline that returns a list of desktop object dicts that the current user has access to.

    :param request: the HTTP request (required).
    :param collection. The Mongo collection name (required).
    :param volume_id: the volume_id of the Mongo database. If None, the root volume is assumed.
    :return: a 200 status code response with the desktop object dicts from the pipeline.
    """
    async with MongoContext(request, volume_id) as mongo:
        sub = request.headers.get(SUB)
        result: list[DesktopObject] = []
        async for r in mongo.aggregate(collection, pipeline):
            obj = heaobject.root.desktop_object_from_dict(r)
            permitted = has_permissions(obj=obj, sub=sub, permissions=RESTPermissionGroup.GETTER_PERMS.perms)
            if permitted:
                result.append(obj)
        return await response.get_all(request, [obj.to_dict() for obj in result])



from jsonschema.validators import Draft7Validator
from jsonschema import ValidationError  # Leave these here so other modules can use them.
from . import jsonschema
from jsonschema import RefResolver
from importlib.resources import as_file, files as resource_container
from . import jsonschemafiles


def compile(schema: str | dict) -> Draft7Validator:
    """
    Returns the correct validator for the given schema.

    :param schema: a JSON schema (required).
    :return: a validator object.
    """
    with as_file(resource_container(jsonschemafiles)) as path:
        resolver = RefResolver(f'{path.as_uri()}/', referrer=schema)
        return Draft7Validator(schema=schema, resolver=resolver)



WSTL_ACTION_SCHEMA_VALIDATOR = compile(jsonschema.WSTL_ACTION_SCHEMA)
WSTL_SCHEMA_VALIDATOR = compile(jsonschema.WSTL_SCHEMA)
CJ_TEMPLATE_SCHEMA_VALIDATOR = compile(jsonschema.CJ_TEMPLATE_SCHEMA)
NVPJSON_SCHEMA_VALIDATOR = compile(jsonschema.NVPJSON_SCHEMA)



from .componenttestcase import ComponentTestCase, ComponentTestCase2, ComponentTestCase3
from heaserver.service.representor import cj
from heaserver.service.testcase.mixin import GetOneMixin
from heaserver.service.oidcclaimhdrs import SUB
from heaobject.user import NONE_USER
from heaobject.volume import MongoDBFileSystem
from aiohttp import hdrs
import aiohttp.client_exceptions


class TestGetComponent(ComponentTestCase, GetOneMixin):  # type: ignore
    async def test_get_by_type(self):
        """
        Checks if a GET request for the object with the expected resources type in ``_expected_one`` succeeds and
        returns the expected data. The test is skipped if the object has no resources or if the type of its resources
        is not defined.
        """
        obj_resources = self._expected_one_wstl[0]['wstl']['data'][0].get('resources', None) or None
        if obj_resources is None:
            self.skipTest('the _expected_one object has no resources')
        obj_resources_type = None
        for resource in obj_resources:
            obj_resources_type = resource.get('resource_type_name', None)
            if obj_resources_type is not None:
                break
        if obj_resources_type is None:
            self.skipTest('the object\'s resources do not have a type')
        response = await self.client.request('GET',
                                             (self._href / 'bytype' / obj_resources_type).path,
                                             headers={**self._headers, hdrs.ACCEPT: cj.MIME_TYPE})
        try:
            self._assert_equal_ordered(self._expected_one[0]['collection']['items'][0]['data'], (await response.json())[0]['collection']['items'][0]['data'])
        except aiohttp.client_exceptions.ContentTypeError as e:
            raise AssertionError(f'GET request failed ({await response.text()}): {e}') from e

    async def test_get_by_type_invalid_type(self):
        """Checks if a GET request for the object with the type 'foo.bar' fails with status 404."""
        response = await self.client.request('GET',
                                             (self._href / 'bytype' / 'foo.bar').path,
                                             headers={**self._headers, hdrs.ACCEPT: cj.MIME_TYPE})
        self.assertEqual(404, response.status)


class TestGetComponent2(ComponentTestCase2, GetOneMixin):  # type: ignore
    pass


class TestGetComponent3(ComponentTestCase3, GetOneMixin):  # type: ignore
    pass


class TestMissingFileSystemAttribute(ComponentTestCase):
    _headers = {SUB: NONE_USER, hdrs.X_FORWARDED_HOST: 'localhost:8080'}

    async def test_missing_file_system_attribute(self):
        obj = await self.client.request('GET',
                                        (self._href / 'bytype' / 'heaobject.folder.Folder' / 'byfilesystemtype' / MongoDBFileSystem.get_type_name()).path,
                                        headers=TestMissingFileSystemAttribute._headers)
        self.assertEqual(200, obj.status)



import asyncio
from http.server import SimpleHTTPRequestHandler
from aiohttp.web import StreamResponse
from aiohttp import StreamReader
from aiohttp.hdrs import CONTENT_DISPOSITION
from aiohttp.test_utils import make_mocked_request
from socketserver import TCPServer, ThreadingMixIn
from threading import Thread
from unittest import IsolatedAsyncioTestCase
from heaserver.service.aiohttp import client_session, StreamReaderWrapper, StreamResponseFileLikeWrapper, RequestFileLikeWrapper, ConnectionFileLikeObjectWrapper
from pathlib import Path
import inspect
import os
from abc import ABC
from unittest import mock
from contextlib import closing


class HTTPServer(ThreadingMixIn, TCPServer):
    pass


class AbstractMyAioHTTPTestCase(IsolatedAsyncioTestCase, ABC):
    def setUp(self) -> None:
        self.old_cwd = os.getcwd()
        filename = inspect.getframeinfo(inspect.currentframe()).filename
        path = os.path.dirname(os.path.abspath(filename))
        os.chdir(path)
        self.data = 'aiohttpdata/requirements_dev.txt'

    def tearDown(self) -> None:
        os.chdir(self.old_cwd)


class MyAioHTTPTestCase(AbstractMyAioHTTPTestCase):

    def setUp(self) -> None:
        super().setUp()
        self.server = HTTPServer(('localhost', 0), SimpleHTTPRequestHandler)
        self.port = self.server.server_address[1]
        self.server_thread = Thread(target=self.server.serve_forever)
        self.server_thread.start()

    async def test_client_session(self):
        async with client_session() as session:
            async with session.get(f'http://localhost:{self.port}') as response:
                self.assertEqual(200, response.status)

    async def test_stream_reader_wrapper(self):

        async with client_session() as session:
            async with session.get(f'http://localhost:{self.port}/{self.data}') as response:
                wrapper = StreamReaderWrapper(response.content)
                try:
                    self.assertEqual(Path(self.data).read_bytes(), await wrapper.read())
                finally:
                    wrapper.close()
                self.assertEqual(200, response.status)

    def tearDown(self) -> None:
        super().tearDown()
        self.server.shutdown()
        self.server.server_close()


class MyAioHTTPTestCase2(AbstractMyAioHTTPTestCase):

    async def test_stream_response_file_like_wrapper(self):
        filename = Path(self.data).name
        response_ = StreamResponse(status=200, reason='OK',
                                   headers={CONTENT_DISPOSITION: f'attachment; filename={filename}'})

        await response_.prepare(make_mocked_request(method='GET', path=f'/{filename}'))
        async with StreamResponseFileLikeWrapper(response_) as fileobj:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, fileobj.write, Path(self.data).read_bytes())

    async def test_request_file_like_wrapper(self):
        loop = asyncio.get_event_loop()
        protocol = mock.Mock(_reading_paused=False)
        filename = Path(self.data).name
        payload = StreamReader(protocol, 2**16, loop=loop)
        payload.feed_data(Path(self.data).read_bytes())
        payload.feed_eof()
        request = make_mocked_request(method='GET', path=f'/{filename}', payload=payload)
        with closing(RequestFileLikeWrapper(request, loop=loop)) as wrapper:
            wrapper.initialize()
            self.assertEquals(Path(self.data).read_bytes(), await loop.run_in_executor(None, wrapper.read))

    async def test_connection_file_like_object_wrapper(self):
        from multiprocessing import Pipe
        a, b = Pipe()
        the_bytes = Path(self.data).read_bytes()
        with closing(ConnectionFileLikeObjectWrapper(a)) as wrapper, closing(b) as b:
            b.send_bytes(the_bytes)
            self.assertEqual(the_bytes, wrapper.read())



"""
Utility classes and functions for working with aiohttp.
"""
import asyncio
from contextlib import AbstractContextManager
from multiprocessing.connection import Connection
from io import BytesIO
from aiohttp.streams import StreamReader
from aiohttp.web import StreamResponse, Request
from aiohttp.client import ClientSession
from typing import Protocol, AsyncIterator, BinaryIO
from heaobject.root import json_dumps
from .appproperty import HEA_COMPONENT
from contextlib import asynccontextmanager
import logging
import os
import platform
import time

from yarl import URL

if platform.system() == 'Windows':
    try:
        import msvcrt
        from ctypes import windll, byref, wintypes, GetLastError, POINTER
        from ctypes.wintypes import HANDLE, BOOL

        LPDWORD = POINTER(wintypes.DWORD)
        PIPE_NOWAIT = wintypes.DWORD(0x00000001)
        ERROR_NO_DATA = 232


        class WindowsNonBlockingFD:
            """
            A wrapper for a file descriptor so that we can call:
                `<AdvancedFD>.read(number_of_bytes)` and
                `<AdvancedFD>.write(data_as_bytes)`

            It also makes the `read_fd` non blocking. When reading from a non-blocking
            pipe with no data it returns b"".

            Methods:
                write(data: bytes) -> None
                read(number_of_bytes: int) -> bytes
                rawfd() -> int
                close() -> None
            """

            def __init__(self, fd: int):
                self.fd = fd
                self.closed = False
                self.config_non_blocking()

            def __del__(self) -> None:
                """
                When this object is garbage collected close the fd
                """
                self.close()

            def close(self) -> None:
                """
                Closes the file descriptor.
                Note: it cannot be reopened and might raise an error if it is
                being used. You don't have to call this function. It is automatically
                called when this object is being garbage collected.
                """
                self.closed = True

            def write(self, data: bytes) -> None:
                """
                Writes a string of bytes to the file descriptor.
                Note: Must be bytes.
                """
                os.write(self.fd, data)

            def read(self, x: int) -> bytes | None:
                """
                Reads `x` bytes from the file descriptor.
                Note: `x` must be an int
                      Returns the bytes. Use `<bytes>.decode()` to convert it to a str
                """
                try:
                    res = os.read(self.fd, x)
                    logging.debug('Result is %s', res)
                    return res
                except OSError as error:
                    err_code = GetLastError()
                    # If the error code is `ERROR_NO_DATA`
                    if err_code == ERROR_NO_DATA:
                        # Return an empty string of bytes
                        logging.debug('Returning None')
                        return None
                    else:
                        logging.debug('Returning error')
                        # Otherwise raise the error
                        website = "https://learn.microsoft.com/en-us/windows/win32/" + \
                                  "debug/system-error-codes--0-499-"
                        raise OSError(
                            f'An exception occurred. Error code: {err_code} Look up the error code here: {website}')

            def config_non_blocking(self) -> bool:
                """
                Makes the file descriptor non blocking.
                Returns `True` if sucessfull, otherwise returns `False`
                """
                # https://stackoverflow.com/a/34504971/11106801
                SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
                SetNamedPipeHandleState.argtypes = [HANDLE, LPDWORD, LPDWORD, LPDWORD]
                SetNamedPipeHandleState.restype = BOOL

                handle = msvcrt.get_osfhandle(self.fd)

                res = windll.kernel32.SetNamedPipeHandleState(handle,
                                                              byref(PIPE_NOWAIT), None,
                                                              None)
                logging.debug('config_non_blocking %s', res)
                return not (res == 0)

            def rawfd(self) -> int:
                """
                Returns the raw fd as an int.
                """
                return self.fd
    except ModuleNotFoundError:
        raise ModuleNotFoundError('Neither set_blocking nor msvcrt found; this module will not work')

CHUNK_SIZE = 16 * 1024


@asynccontextmanager
async def client_session(*args, **kwargs) -> AsyncIterator[ClientSession]:
    """
    Gets a new aiohttp ClientSession. The session's json_serialize parameter is set to heaobject.root.json_dumps,
    which unlike Python's native json serialization supports serializing timestamps. It supports any arguments that
    ClientSession's constructor supports.
    """
    session = ClientSession(json_serialize=json_dumps, *args, **kwargs)
    try:
        yield session
    finally:
        await session.close()


class ConnectionFileLikeObjectWrapper(AbstractContextManager):
    """
    Wraps a multiprocessing.connection.Connection object and provides file-like object methods.

    This class is a context manager, so it can be used in with statements.
    """

    def __init__(self, conn: Connection):
        """
        Creates a new ConnectionFileLikeObjectWrapper object, passing in the connection to wrap.

        :param conn: a multiprocessing.connection.Connection object (required).
        """
        if conn is None:
            raise ValueError('conn cannot be None')
        self.__conn = conn
        self.__buffer = BytesIO()

    def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        if len(b := self.__buffer.read(n)) > 0:
            return b
        try:
            result: bytes = self.__conn.recv_bytes()
            if -1 < n < len(result):
                pos = self.__buffer.tell()
                self.__buffer.write(result[n:])
                self.__buffer.seek(pos)
                return result[:n]
            else:
                return result
        except EOFError:
            return b''

    def write(self, b):
        """
        Sends some bytes to the connection.

        :param b: some bytes (required).
        """
        self.__conn.send_bytes(b)

    def fileno(self):
        """
        Returns the integer file descriptor that is used by the connection.

        :return: the integer file descriptor.
        """
        return self.__conn.fileno()

    def close(self):
        """
        Closes the connection and any other resources associated with this object.
        """
        try:
            self.__buffer.close()
            self.__conn.close()
            self.__conn = None
        finally:
            if self.__conn is not None:
                try:
                    self.conn.close()
                except OSError:
                    pass

    def __exit__(self, *exc_details):
        self.close()


class SupportsAsyncRead(Protocol):
    """
    Protocol with an async read() method and a close() method.
    """

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        pass

    def close(self):
        """
        Closes any resources associated with this object.
        """
        pass


class AsyncReader:
    """
    Wraps a bytes object in a simple reader with an asynchronous read method and a close method.
    """

    def __init__(self, b: bytes):
        """
        Creates a new AsyncReader, passing in a bytes object.

        :param b: bytes (required).
        """
        self.__b = BytesIO(b)

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return self.__b.read(n)

    def close(self):
        """
        Closes any resources associated with this object.
        """
        self.__b.close()


class StreamReaderWrapper:
    """
    Wraps an aiohttp StreamReader in an asyncio StreamReader-like object with a read() method and a no-op close()
    method.
    """

    def __init__(self, reader: StreamReader):
        if reader is None:
            raise ValueError('reader cannot be None')
        self.__reader = reader

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return await self.__reader.read(n)

    def close(self):
        pass


class StreamResponseFileLikeWrapper:
    """
    Wraps an aiohttp StreamResponse in a file-like object wrapper. The write() method must be called in a separate
    thread.
    """

    def __init__(self, resp: StreamResponse, loop: asyncio.AbstractEventLoop = None):
        if loop is not None:
            self.loop = loop
        else:
            self.loop = asyncio.get_running_loop()
        self.resp = resp
        self.__successful = False
        self.__error_msg = None
        self.__logger = logging.getLogger(__name__)
        read_fd, write_fd = os.pipe()
        self.__logger.debug('File handles %d and %d', read_fd, write_fd)
        self.__writer = os.fdopen(write_fd, 'wb')
        # https://stackoverflow.com/questions/34504970/non-blocking-read-on-os-pipe-on-windows
        if hasattr(os, 'set_blocking'):
            os.set_blocking(read_fd, False)  # type:ignore  # not available on Windows
            self.__reader: BinaryIO | WindowsNonBlockingFD = os.fdopen(read_fd, 'rb')
        elif platform.system() == 'Windows':
            self.__logger.debug('On Windows')
            self.__reader = WindowsNonBlockingFD(read_fd)

        self.__task = self.loop.create_task(self.__pump_bytes_into_fd())

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def write(self, data: bytes) -> int:
        """
        Writes to the StreamResponse and returns how many bytes were written.

        :param data: the bytes to write (required).
        :return: the number of bytes written.
        """
        self.__logger.debug('Writing %s', data)
        return self.__writer.write(data)

    def flush(self):
        """
        Flush the write buffers of the stream if applicable.
        """
        self.__writer.flush()

    async def close(self):
        """
        Cleans up all resources in this file-like object.
        """
        self.__logger.debug('Closing')
        self.__writer.close()  # Triggers a CancelledError in __pump_bytes_into_fd on Windows.
        self.__logger.debug('Writer is closed')
        await self.__task
        self.__logger.debug('Task done')
        if not self.__successful:
            raise IOError(f'Failed writing data to the response: {self.__error_msg}')
        self.__logger.debug('Closed')

    async def __pump_bytes_into_fd(self):
        writer_closed = False
        try:
            self.__logger.debug('About to write')
            while True:
                try:
                    while (chunk := await self.loop.run_in_executor(None, self.__reader.read, CHUNK_SIZE)) != b'':
                        if chunk is None:
                            self.__logger.debug('Did not get anything; will busy wait and try again')
                            await asyncio.sleep(0.1)
                            self.__logger.debug('Trying again')
                            continue
                        self.__logger.debug('Read %d bytes from download', len(chunk))
                        await self.resp.write(chunk)
                        self.__logger.debug('Wrote %d bytes to stream writer', len(chunk))
                    break
                except asyncio.CancelledError as e:
                    # On Windows, closing the write side of the pipe triggers a CancelledError, which we need to deal
                    # with here.
                    self.__logger.debug(f'Cancelled but continuing: {e}')
                    continue
            self.__reader.close()
            writer_closed = True
            self.__successful = True
            self.__logger.debug('Done reading file')
        except OSError as e:
            self.__logger.exception('Failed to read file')
            self.__error_msg = str(e)
            if not writer_closed:
                self.__logger.debug('Closing writer after error')
                try:
                    self.__writer.close()
                except Exception:
                    pass
            if not self.__reader.closed:
                self.__logger.debug('Closing reader after error')
                try:
                    self.__reader.close()
                except:
                    pass
            raise e
        self.__logger.debug('Exiting task')


class RequestFileLikeWrapper:
    """
    Wraps an aiohttp request's content in a file-like object with read() and close() functions. Before doing any
    reading, call the initialize() method. The read() method must be called in a separate thread.
    """

    def __init__(self, request: Request, loop: asyncio.AbstractEventLoop = None):
        """
        Creates the file-like object wrapper.

        :param request: the aiohttp request (required).
        :param loop: the current event loop. If None, it will use asyncio.get_running_loop().
        """
        if loop is not None:
            self.loop = loop
        else:
            self.loop = asyncio.get_running_loop()
        self.request = request
        self.__pump_task: asyncio.Task | None = None
        self.__reader: BinaryIO | None = None
        self.__writer: BinaryIO | None = None

    def initialize(self):
        """
        Creates resources needed for reading.
        """
        read_fd, write_fd = os.pipe()
        self.__pump_task = self.loop.create_task(self.__pump_bytes_into_fd())
        self.__reader = os.fdopen(read_fd, 'rb')
        self.__writer = os.fdopen(write_fd, 'wb')

    def read(self, n=-1, /) -> bytes:
        """
        Reads some bytes.

        :param n: the number of bytes to read (or -1 for everything).
        :return: the bytes that were read.
        """
        logger = logging.getLogger(__name__)
        logger.debug('Reading %s bytes', n)
        output = self.__reader.read(n)
        logger.debug('Read %s bytes', len(output))
        return output

    def close(self):
        """
        Cleans up all resources in this file-like object.
        """
        logger = logging.getLogger(__name__)
        logger.debug('Closing')
        try:
            if self.__writer is not None:
                self.__writer.close()
                self.__writer = None
            if self.__reader is not None:
                self.__reader.close()
                self.__reader = None
            while self.__pump_task is not None and not self.__pump_task.done():
                time.sleep(0.1)
            self.__pump_task = None
        except Exception as e:
            logger.exception('Failed to close pipe')
            if self.__reader is not None:
                try:
                    self.__reader.close()
                    self.__reader = None
                except Exception:
                    pass
            try:
                while self.__pump_task is not None and not self.__pump_task.done():
                    time.sleep(0.1)
                self.__pump_task = None
            except Exception:
                pass
            raise e

    async def __pump_bytes_into_fd(self):
        logger = logging.getLogger(__name__)
        writer_closed = False
        content_type = self.request.headers.get('Content-Type', '')
        try:
            if content_type.startswith('multipart/form-data'):
                logger.debug(f'Is multipart')
                mp_reader = await self.request.multipart()
                while True:
                    part = await mp_reader.next()
                    if part is None:
                        break
                    logger.debug("the part's headers %s" % part.headers)
                    while True:
                        chunk = await part.read_chunk(CHUNK_SIZE)
                        if not chunk or chunk == b'':
                            break
                        logger.debug('Read %d bytes from upload', len(chunk))
                        bytes_written = await self.loop.run_in_executor(None, self.__writer.write, chunk)
                        logger.debug('Wrote %d bytes to pipe', bytes_written)
            else:
                while not self.request.content.at_eof():
                    chunk = await self.request.content.read(CHUNK_SIZE)
                    logger.debug('Read %d bytes from upload', len(chunk))
                    bytes_written = await self.loop.run_in_executor(None, self.__writer.write, chunk)
                    logger.debug('Wrote %d bytes to pipe', bytes_written)
            self.__writer.close()
            writer_closed = True
            logger.debug('Done reading file')
        except Exception as e:
            logger.exception('Failed to read file')
            if not writer_closed:
                try:
                    self.__writer.close()
                except Exception:
                    pass
                writer_closed = False
                raise e

def absolute_url_minus_base(request: Request, url: URL | str | None = None) -> str:
    """
    Removes the current service's base URL from the beginning of the provided URL. If no URL is provided, the request
    URL is used. If the provided URL is not prefixed with the service's base URL, the URL is returned unaltered.

    :param request: the HTTP request (required).
    :param url: optional URL.
    :return: the resulting URL.
    """
    if url is None:
        url_ = str(request.url)
    else:
        url_ = str(url)
    base_url = request.app[HEA_COMPONENT]
    if not url_.startswith(base_url):
        return url
    else:
        return url_.removeprefix(base_url).lstrip('/')
